var
  fs = require("fs"),
  path = require("path"),
  _ = require("lodash"),
  async = require("async"),
  ProgressBar = require("progress"),
  http = require("http"),
  unzip = require("unzip"),
  spawn = require("child_process").spawn,
  jarfile = path.normalize(__dirname + "/Sauce-Connect.jar"),
  zipfile = path.normalize(__dirname + "/Sauce-Connect-latest.zip"),
  exists = fs.existsSync || path.existsSync,
  openProcesses = [],
  logger = console.log,
  cleanup_registered = false;

function killProcesses() {
  _.each(openProcesses, function (proc) {
    proc.emit("exit");
    proc.kill("SIGTERM");
  });
  openProcesses = [];
}

// Make sure all processes have been closed
// when the script goes down
function closeOnProcessTermination() {
  if (cleanup_registered) {
    return;
  }
  cleanup_registered = true;
  process.on("exit", function () {
    logger("Shutting down");
    killProcesses();
  });
}


function download(options, callback) {
  var req = http.request({
      host: "saucelabs.com",
      port: 80,
      path: "/downloads/Sauce-Connect-latest.zip"
    });

  logger("Missing Sauce Connect local proxy, downloading dependency");
  logger("This will only happen once.");

  req.on("response", function (res) {
    var bar, len = parseInt(res.headers["content-length"], 10),
      prettyLen = (len / (1024 * 1024) + "").substr(0, 4);

    logger("Downloading ", prettyLen, " MB");


    res.pipe(fs.createWriteStream(zipfile));

    if (!options.no_progress) {
      bar = new ProgressBar("  downloading Sauce-Connect-latest.zip [:bar] :percent :etas", {
        complete: "=",
        incomplete: " ",
        width: 20,
        total: len
      });

      res.on("data", function (chunk) {
        bar.tick(chunk.length);
      });
    }

    res.on("end", function () {
      logger("\n");
      logger("Unzipping Sauce-Connect-latest.zip");
      setTimeout(function () {
        fs.createReadStream(zipfile)
          .pipe(unzip.Parse())
          .on("entry", function (entry) {
            if (entry.path.match(/Sauce-Connect\.jar$/)) {
              logger("Saving Sauce-Connect.jar");
              var jar = fs.createWriteStream(jarfile);
              entry.on("end", function () {
                // write queued data before closing the stream
                logger("Removing Sauce-Connect-latest.zip");
                fs.unlink(zipfile, function () {
                  logger("Sauce Connect installed correctly");
                  callback(null);
                });
              });
              entry.pipe(jar);
            }
          });
      }, 1000);

    });

  });

  req.end();
}



function run(options, callback) {
  callback = _.once(callback);
  logger("Opening local tunnel using Sauce Connect");
  var child,
    args = ["-jar", jarfile, options.username, options.accessKey],
    dataActions = {
      "Please wait for 'You may start your tests' to start your tests": function connecting() {
        logger("Creating tunnel with Sauce Labs");
      },
      "Connected! You may start your tests": function ready() {
        logger("Testing tunnel ready");
        closeOnProcessTermination();
        callback(null, child);
      },
      "This version of Sauce Connect is outdated": function outdated() {

      },
      "Exception: ": function error(data) {
        if (data.indexOf("HTTP response code: 401") !== -1) {
          logger("Invalid Sauce Connect Credentials");
          return callback(new Error("Invalid Sauce Connect Credentials. " + data), child);
        }
        logger("Sauce Connect Error");
        callback(new Error(data), child);
      },
      "Finished shutting down tunnel remote VM: ": function shutDown() {

      }
    };

  if (options.port) {
    args.push("-P", options.port);
  }

  if (options.logfile) {
    args.push("-l", options.logfile);
  }

  if (options.tunnelIdentifier) {
    args.push("--tunnel-identifier", options.tunnelIdentifier);
  }

  child = spawn("java", args);

  child.stdout.on("data", function (data) {
    data = data.toString();
    if (options.verbose) {
      console.log(data);
    }

    _.each(dataActions, function (action, key) {
      if (data.indexOf(key) !== -1) {
        action(data);
        return false;
      }
    });
  });

  child.on("exit", function (code, signal) {
    if (code > 0) {
      callback(new Error("Could not start Sauce Connect. Exit code " + code +
        " signal: " + signal));
    }
  });

  openProcesses.push(child);

  child.close = function () {
    child.kill("SIGTERM");
  };
}

function downloadAndStartProcess(options, callback) {
  if (arguments.length === 1) {
    callback = options;
    options = {};
  }
  logger = options.logger || function () {};

  function checkForZip(next) {
    if (!exists(zipfile)) {
      download(options, next);
    } else {
      // the zip is being downloaded, poll for the jar to be ready
      async.doUntil(async.apply(_.delay, 1000), async.apply(exists, jarfile), next);
    }
  }

  async.waterfall([
    function checkForJar(next) {
      if (exists(jarfile)) {
        next(null);
      } else {
        checkForZip(next);
      }
    },
    async.apply(run, options)
  ], callback);

}



module.exports = downloadAndStartProcess;
module.exports.kill = killProcesses;


